//==============================================================================
// app.js
// Copyright (c) 2015 - 2020 Douraku
// Released under the MIT License.
// http://opensource.org/licenses/mit-license.php
//==============================================================================

/*:
 * @target MZ
 * @plugindesc アプリケーション専用プラグイン ver1.0.0
 * @author 道楽
 * 
 * @requiredAssets	img/system/Window_Message.png
 *
 * @command GainItem
 * @text アイテム獲得
 * @desc 変数で指定したアイテム/武器/防具を獲得する
 *
 * @arg itemType
 * @type combo
 * @default item
 * @option weapon
 * @option armor
 * @text アイテム種類
 * @desc 獲得するアイテムの種類を選択する
 *
 * @arg dataVariableId
 * @type number
 * @default 1
 * @text データＩＤの変数番号
 * @desc 変数の管理番号を設定する
 *
 * @arg amountVariableId
 * @type number
 * @default 0
 * @text 獲得数の変数番号
 * @desc 変数の管理番号を設定する
 *
 * @command CompleteItem
 * @text 全アイテム獲得
 * @desc すべてのアイテムを最大数まで獲得する
 *
 * @command SimpleButtonPicture
 * @text ピクチャボタン拡張を有効化
 * @desc ピクチャボタンにOnPressで画像が変化する仕組みを追加する
 *
 * @arg pictureId
 * @type number
 * @min 1
 * @max 100
 * @default 1
 * @text ピクチャ番号
 * @desc ピクチャの管理番号
 *
 * @arg overwriteOpacity
 * @type boolean
 * @default true
 * @text 透明度上書きフラグ
 * @desc 透明度を上書きするか
 */

var Imported = Imported || {};
Imported.app = true;

(function (exports) {
	'use strict';

	//--------------------------------------------------------------------------
	/** PluginManager */
	const pluginName = "app";

	PluginManager.registerCommand(pluginName, "GainItem", args => {
		const vIdData = Number(args.dataVariableId);
		const vIdAmount = Number(args.amountVariableId);
		const dataId = $gameVariables.value(vIdData);
		const amount = $gameVariables.value(vIdAmount);
		switch (args.itemType)
		{
		case "item":   $gameParty.gainItem($dataItems[dataId], amount);   break;
		case "weapon": $gameParty.gainItem($dataWeapons[dataId], amount); break;
		case "armor":  $gameParty.gainItem($dataArmors[dataId], amount);  break;
		}
	});

	PluginManager.registerCommand(pluginName, "CompleteItem", args => {
		$dataItems.forEach(item => {
			if ( item && item.name.length > 0 && !item.meta.ignoreItemComplete )
			{
				$gameParty.gainItem(item, $gameParty.maxItems(item));
			}
		});
		$dataWeapons.forEach(item => {
			if ( item && item.name.length > 0 && !item.meta.ignoreItemComplete )
			{
				$gameParty.gainItem(item, $gameParty.maxItems(item));
			}
		});
		$dataArmors.forEach(item => {
			if ( item && item.name.length > 0 && !item.meta.ignoreItemComplete )
			{
				$gameParty.gainItem(item, $gameParty.maxItems(item));
			}
		});
	});

	PluginManager.registerCommand(pluginName, "SimpleButtonPicture", args => {
		const pictureId = Number(args.pictureId);
		const overwriteOpacity = Boolean(args.overwriteOpacity === "true" ? true : false);
		const picture = $gameScreen.picture(pictureId);
		if (picture)
		{
			picture.setButtonPicture(true, overwriteOpacity);
		}
	});

	//--------------------------------------------------------------------------
	/** ColorManager */
	ColorManager.castGaugeColor1 = function() { return this.textColor(14); };
	ColorManager.castGaugeColor2 = function() { return this.textColor(6); };
	ColorManager.messageColor = function() { return this.textColor(15); };
	ColorManager.messageOutlineColor = function() { return "rgba(255, 255, 255, 0.6)"; };

	//--------------------------------------------------------------------------
	/** TextManager */
	TextManager._scopeText =
	[
		"",
		"敵単体",
		"敵全体",
		"敵ランダム１体",
		"敵ランダム２体",
		"敵ランダム３体",
		"敵ランダム４体",
		"味方単体",
		"味方全体",
		"味方単体",
		"味方全体",
		"使用者",
		"味方単体",
		"味方全体",
		"敵味方全体",
	];
	TextManager._skillCategoryText =
	[
		"",
		"\\c[2]【攻撃】",
		"\\c[6]【防御】",
		"\\c[3]【回復】",
		"\\c[1]【補助】",
	];
	TextManager._elementText =
	[
		"\\c[7]",
		"\\c[0]",
		"\\c[18]",
		"\\c[22]",
		"\\c[17]",
		"\\c[0]",
		"\\c[0]",
		"\\c[0]",
	];

	TextManager.scope = function(scope)
	{
		return this._scopeText[scope];
	};

	TextManager.skillCategory = function(category)
	{
		return this._skillCategoryText[category];
	};

	TextManager.element = function(elementId)
	{
		var element = $dataSystem.elements[elementId];
		if ( elementId === 0 )
		{
			element = "-";
		}
		return this._elementText[elementId] + element;
	};

	//--------------------------------------------------------------------------
	/** Game_Screen */
	(function () {
		const base = Game_Screen.prototype.clear;
		Game_Screen.prototype.clear = function()
		{
			base.call(this);
			this.clearQuake();
		};
	}());

	(function () {
		const base = Game_Screen.prototype.updateShake;
		Game_Screen.prototype.updateShake = function()
		{
			base.call(this);
			this.updateQuake();
		};
	}());

	Game_Screen.prototype.clearQuake = function()
	{
		this._quakeX = { value: 0, direction: +1, power: 2 };
		this._quakeY = { value: 0, direction: -1, power: 2 };
	};

	Game_Screen.prototype.updateQuake = function()
	{
		if ( !$gameSwitches.value(42) && this.quakeX() === 0 && this.quakeY() === 0 )
		{
			this.clearQuake();
			return;
		}
		const func = (data) =>
		{
			if ( data.direction === 0 )
			{
				return;
			}
			const delta = data.direction * data.power;
			const nextValue = data.value + delta;
			data.value = (data.value * nextValue < 0) ? 0 : nextValue;
			let nextDirection = data.direction;
			if ( data.value > +data.power * 2 )
			{
				nextDirection = -1;
			}
			if ( data.value < -data.power * 2 )
			{
				nextDirection = +1;
			}
			if ( data.direction * nextDirection < 0 )
			{
				data.direction = nextDirection;
				data.power = Math.random(10) + 1;
			}
		};
		func(this._quakeX);
		func(this._quakeY);
	};

	Game_Screen.prototype.quakeX = function() { return this._quakeX.value; };
	Game_Screen.prototype.quakeY = function() { return this._quakeY.value; };

	//--------------------------------------------------------------------------
	/** Game_Battler */
	Game_Battler.prototype.isTpbCasting = function()
	{
		return this._tpbState === "casting";
	};

	Game_Battler.prototype.tpbCastTime = function()
	{
		return this._tpbCastTime;
	};

	//--------------------------------------------------------------------------
	/** Game_Actor */
	(function () {
		const base = Game_Actor.prototype.skills;
		Game_Actor.prototype.skills = function()
		{
			var list = base.call(this);
			list.sort((a, b) => a.id - b.id);
			return list;
		};
	}());

	Game_Actor.prototype.optimizeEquipments = function()
	{
		const maxSlots = this.equipSlots().length;
		this.clearOptimizeEquipments();
		for ( let i = 0; i < maxSlots; i++ )
		{
			if ( this.isEquipOptimizeOk(i) )
			{
				this.changeEquip(i, this.bestEquipItem(i));
			}
		}
	};

	Game_Actor.prototype.clearOptimizeEquipments = function()
	{
		const maxSlots = this.equipSlots().length;
		for (let i = 0; i < maxSlots; i++)
		{
			if ( this.isEquipOptimizeOk(i) )
			{
				this.changeEquip(i, null);
			}
		}
	};

	Game_Actor.prototype.isEquipOptimizeOk = function(slotId)
	{
		if ( !this.isEquipChangeOk(slotId) )
		{
			return false;
		}
		return (slotId !== 4); // 装飾品は含まない
	};

	//--------------------------------------------------------------------------
	/** Game_Player */
	Game_Player.prototype.existFollower = function(index)
	{
		return this._followers.follower(index) ? true : false;
	};

	Game_Player.prototype.resetPatternWithFollower = function()
	{
		this.resetPattern();
		this._followers.resetPattern();
	};

	Game_Player.prototype.setFollowerDirection = function(index, d)
	{
		if ( this.existFollower(index) )
		{
			this._followers.follower(index).setDirection(d);
		}
	};

	Game_Player.prototype.findMemberIndexByActorId = function(actorId)
	{
		return this.members().indexOf(actorId);
	};

	//--------------------------------------------------------------------------
	/** Game_Followers */
	Game_Followers.prototype.resetPattern = function()
	{
		for (const follower of this._data)
		{
			follower.resetPattern();
		}
	};

	//--------------------------------------------------------------------------
	/** Game_Picture */
	(function () {
		const base = Game_Picture.prototype.initBasic;
		Game_Picture.prototype.initBasic = function(pictureId)
		{
			base.apply(this, arguments);
			this._buttonPicture = false;
			this._overwriteOpacity = false;
		};
	}());

	Game_Picture.prototype.setButtonPicture = function(enable, overwriteOpacity)
	{
		this._buttonPicture = enable;
		this._overwriteOpacity = overwriteOpacity;
	};

	Game_Picture.prototype.buttonPicture = function() { return this._buttonPicture; };
	Game_Picture.prototype.overwriteOpacity = function() { return this._overwriteOpacity; };

	//--------------------------------------------------------------------------
	/** Sprite_Picture */
	(function () {
		const base = Sprite_Picture.prototype.initialize;
		Sprite_Picture.prototype.initialize = function(pictureId)
		{
			this._buttonPicture = false;
			this._overwriteOpacity = false;
			this._coldFrame = null;
			this._hotFrame = null;
			base.apply(this, arguments);
		};
	}());

	(function () {
		const base = Sprite_Picture.prototype.update;
		Sprite_Picture.prototype.update = function()
		{
			base.apply(this, arguments);
			if(this._buttonPicture && !$gameMessage.isBusy())
			{
				this.updateFrame();
				this.updateOpacity();
			}
		};
	}());

	(function () {
		const base = Sprite_Picture.prototype.loadBitmap;
		Sprite_Picture.prototype.loadBitmap = function()
		{
			base.apply(this, arguments);
			this.bitmap.addLoadListener(() => {
				const picture = this.picture();
				if (!picture)
				{
					return;
				}
				this._buttonPicture = picture.buttonPicture();
				this._overwriteOpacity = picture.overwriteOpacity();
				if (this._buttonPicture)
				{
					const width  = this.bitmap.width;
					const height = this.bitmap.height / 2;
					this._coldFrame = new Rectangle(0, 0, width, height);
					this._hotFrame = new Rectangle(0, height, width, height);
				}
			});
		};
	}());

	Sprite_Picture.prototype.updateFrame = function()
	{
		const frame = (this.isPressed() && this.isClickEnabled()) ? this._hotFrame : this._coldFrame;
		if (frame)
		{
			this.setFrame(frame.x, frame.y, frame.width, frame.height);
		}
	};

	Sprite_Picture.prototype.updateOpacity = function()
	{
		if (this._overwriteOpacity)
		{
			this.opacity = this.isPressed() ? 255 : 192;
		}
	};

	//--------------------------------------------------------------------------
	/** Sprite_Gauge */
	(function () {
		const base = Sprite_Gauge.prototype.currentValue;
		Sprite_Gauge.prototype.currentValue = function()
		{
			return this.isCastingTimeGauge() ? this._battler.tpbCastTime() : base.apply(this, arguments);
		};
	}());

	(function () {
		const base = Sprite_Gauge.prototype.currentMaxValue;
		Sprite_Gauge.prototype.currentMaxValue = function()
		{
			return this.isCastingTimeGauge() ? this._battler.tpbRequiredCastTime() : base.apply(this, arguments);
		};
	}());

	(function () {
		const base = Sprite_Gauge.prototype.gaugeColor1;
		Sprite_Gauge.prototype.gaugeColor1 = function()
		{
			return this.isCastingTimeGauge() ? ColorManager.castGaugeColor1() : base.apply(this, arguments);
		};
	}());

	(function () {
		const base = Sprite_Gauge.prototype.gaugeColor2;
		Sprite_Gauge.prototype.gaugeColor2 = function()
		{
			return this.isCastingTimeGauge() ? ColorManager.castGaugeColor2() : base.apply(this, arguments);
		};
	}());

	Sprite_Gauge.prototype.isCastingTimeGauge = function()
	{
		if ( this._statusType !== "time" )
		{
			return false;
		}
		if ( !this._battler )
		{
			return false;
		}
		if ( !this._battler.isTpbCasting() )
		{
			return false;
		}
		if ( this._battler.tpbRequiredCastTime() <= 0 )
		{
			return false;
		}
		return true;
	};

	//--------------------------------------------------------------------------
	/** Sprite_Character */
	(function () {
		const base = Sprite_Character.prototype.setCharacterBitmap;
		Sprite_Character.prototype.setCharacterBitmap = function()
		{
			base.apply(this, arguments);
			this.bitmap.addLoadListener(bitmap => { bitmap.smooth = false; });
		};
	}());

	//--------------------------------------------------------------------------
	/** Spriteset_Base */
	(function () {
		const base = Spriteset_Base.prototype.updatePosition;
		Spriteset_Base.prototype.updatePosition = function()
		{
			base.call(this);
			this.x += Math.round($gameScreen.quakeX());
			this.y += Math.round($gameScreen.quakeY());
		};
	}());

	//--------------------------------------------------------------------------
	/** Window_Base */
	(function () {
		const base = Window_Base.prototype.convertEscapeCharacters;
		Window_Base.prototype.convertEscapeCharacters = function(text)
		{
			text = text.replace(/\\/g, "\x1b");
			text = text.replace(/\x1b\x1b/g, "\\");
			text = text.replace(/\x1bWN\[(\d+)\]/gi, (_, p1) =>
				this.weaponName(parseInt(p1))
			);
			text = text.replace(/\x1bAN\[(\d+)\]/gi, (_, p1) =>
				this.armorName(parseInt(p1))
			);
			text = text.replace(/\x1bIN\[(\d+)\]/gi, (_, p1) =>
				this.itemName(parseInt(p1))
			);
			text = text.replace(/\x1bSN\[(\d+)\]/gi, (_, p1) =>
				this.skillName(parseInt(p1))
			);
			text = text.replace(/\x1bWD\[(\d+)\]/gi, (_, p1) =>
				this.weaponDescription(parseInt(p1))
			);
			text = text.replace(/\x1bAD\[(\d+)\]/gi, (_, p1) =>
				this.armorDescription(parseInt(p1))
			);
			text = text.replace(/\x1bSD\[(\d+)\]/gi, (_, p1) =>
				this.skillDescription(parseInt(p1))
			);
			return base.call(this, text);
		};
	}());

	Window_Base.prototype.weaponName = function(itemId)
	{
		return this._implItemNameText($dataWeapons[itemId]);
	};

	Window_Base.prototype.armorName = function(itemId)
	{
		return this._implItemNameText($dataArmors[itemId]);
	};

	Window_Base.prototype.itemName = function(itemId)
	{
		return this._implItemNameText($dataItems[itemId]);
	};

	Window_Base.prototype.skillName = function(itemId)
	{
		return this._implItemNameText($dataSkills[itemId]);
	};

	Window_Base.prototype._implItemNameText = function(item)
	{
		var text = "\\i[";
		text += item.iconIndex;
		text += "]";
		text += item.name;
		return text;
	};

	Window_Base.prototype.weaponDescription = function(weaponId)
	{
		const item = $dataWeapons[weaponId];
		var text = "\\c[2]【" + $dataSystem.equipTypes[item.etypeId] + "】";
		//text += "\\c[8]分類:\\c[0]" + $dataSystem.weaponTypes[item.wtypeId];
		text += this._implItemParamText(item);
		text += "\\c[0]";
		return text;
	};

	Window_Base.prototype.armorDescription = function(armorId)
	{
		const item = $dataArmors[armorId];
		var text = "\\c[2]【" + $dataSystem.equipTypes[item.etypeId] + "】";
		//text += "\\c[8]分類:\\c[0]" + $dataSystem.armorTypes[item.atypeId];
		text += this._implItemParamText(item);
		text += "\\c[0]";
		return text;
	};

	Window_Base.prototype.skillDescription = function(skillId)
	{
		const item = $dataSkills[skillId];
		var category = this._skillCategory(item);
		var text = TextManager.skillCategory(category);
		if ( item.scope > 0 )
		{
			text += "\\c[8]範囲:\\c[0]" + TextManager.scope(item.scope);
		}
		if ( category === 1 ) // 攻撃
		{
			if ( item.damage.elementId >= 0 )
			{
				text += " \\c[8]属性:" + TextManager.element(item.damage.elementId);
			}
		}
		text += "\\c[0]";
		return text;
	};

	Window_Base.prototype._implItemParamText = function(item)
	{
		var text = "";
		for ( let ii = 0; ii < 8; ii++ )
		{
			if ( item.params[ii] !== 0 )
			{
				if ( text !== "" )
				{
					text += " ";
				}
				text += "\\c[8]" + $dataSystem.terms.params[ii] + ":\\c[0]";
				text += item.params[ii];
			}
		}
		for ( let ii = 0; ii < 2; ii++ )
		{
			var xparam = this.xparam(item, ii);
			if ( xparam !== 0 )
			{
				if ( text !== "" )
				{
					text += " ";
				}
				text += "\\c[8]" + $dataSystem.terms.params[8+ii] + ":\\c[0]";
				text += ((xparam > 0) ? "+" : "") + xparam + "%";
			}
		}
		return text;
	};

	Window_Base.prototype.traitsWithId = function(item, code, id)
	{
		return item.traits.filter(
			trait => trait.code === code && trait.dataId === id
		);
	};

	Window_Base.prototype.traitsSum = function(item, code, id)
	{
		return this.traitsWithId(item, code, id).reduce((r, trait) => r + trait.value, 0);
	};

	Window_Base.prototype.xparam = function(item, xparamId)
	{
		return Math.floor(this.traitsSum(item, Game_BattlerBase.TRAIT_XPARAM, xparamId) * 100);
	};

	Window_Base.prototype._skillCategory = function(item)
	{
		return item.meta.category ? Number(item.meta.category) : 0;
	};

	//--------------------------------------------------------------------------
	/** Window_MapName */
	Window_MapName.prototype.refresh = function()
	{
		this.contents.clear();
		if ( $gameMap.displayName() )
		{
			const width = this.innerWidth;
			const text = this.convertEscapeCharacters($gameMap.displayName());
			this.drawBackground(0, 0, width, this.lineHeight());
			this.drawText(text, 0, 0, width, "center");
		}
	};

	//--------------------------------------------------------------------------
	/** Window_Help */
	(function () {
		const base = Window_Help.prototype.setItem;
		Window_Help.prototype.setItem = function(item)
		{
			if ( item )
			{
				var text = item.description;
				if ( DataManager.isWeapon(item) )
				{
					text = "\\wd[" + item.id + "]\n" + text;
				}
				else if ( DataManager.isArmor(item) )
				{
					text = "\\ad[" + item.id + "]\n" + text;
				}
				else if ( DataManager.isSkill(item) )
				{
					text = "\\sd[" + item.id + "]" + text;
				}
				this.setText(text);
				return;
			}
			base.apply(this, arguments);
		};
	}());

	//--------------------------------------------------------------------------
	/** Window_Message */
	Window_Message.prototype.updateBackOpacity = function()
	{
		this.backOpacity = 255;
	};

	(function () {
		const base = Window_Message.prototype.resetTextColor;
		Window_Message.prototype.resetTextColor = function()
		{
			if ( this.isPopup() )
			{
				this.changeTextColor(ColorManager.messageColor());
				this.changeOutlineColor(ColorManager.messageOutlineColor());
			}
			else
			{
				base.apply(this, arguments);
			}
		};
	}());

	//--------------------------------------------------------------------------
	/** Window_ActorCommand */
	Window_ActorCommand.prototype.addSkillCommands = function()
	{
		const skillTypes = this._actor.skillTypes();
		for (const stypeId of skillTypes)
		{
			if ( stypeId === 2 )
			{
				continue; // 能力強化は戦闘中表示しない
			}
			if ( $gameSwitches.value(60) )
			{
				if ( stypeId === 1 )
				{
					continue;
				}
			}
			const name = $dataSystem.skillTypes[stypeId];
			this.addCommand(name, "skill", true, stypeId);
		}
	};

	//--------------------------------------------------------------------------
	/** Scene_Map */
	(function () {
		const base = Scene_Map.prototype.shouldAutosave;
		Scene_Map.prototype.shouldAutosave = function()
		{
			if ( $gameSwitches.value(46) ) // AutoSave禁止スイッチ
			{
				return false;
			}
			return base.apply(this, arguments);
		};
	}());

	//--------------------------------------------------------------------------
	/** Scene_Battle */
	(function () {
		const base = Scene_Battle.prototype.shouldAutosave;
		Scene_Battle.prototype.shouldAutosave = function()
		{
			if ( $gameSwitches.value(46) ) // AutoSave禁止スイッチ
			{
				return false;
			}
			return base.apply(this, arguments);
		};
	}());

	Scene_Battle.prototype.statusWindowX = function()
	{
		return this.statusWindowRect().x;
	};

	//import "Managers/BattleManager.js"

}((this.app = this.app || {})));
